// This only implements the basic widget functions, to make life easier.
// Code to figure out which window/widget was clicked goes here. No drawing code here.

#include "../openttd.h"
#include "new_widgets.h"

/////////////////////////////////////////////////////////////////////////////
// Widget management.
//

// Delete all children of a given widget, and after that delete the widget itself.
inline void DeleteWidget (Widget* w)
{
	static Event cleanup = {E_WIDGET_DESTROY,0,NULL};				// Only matters here.
	Widget* tmp, work;

	assert (w != NULL);
// First get rid of all children.
	tmp = w->child_list;
	if (tmp != NULL)
		do {
			work = tmp;
			tmp = work->next;
			DeleteWidget (work);
		} while (tmp != NULL);
// Clean up the widget itself.
	if (w->on_event != NULL)
		w->on_event (cleanup);
	free (w);
	w = NULL;
}

static inline Widget* BuildWidget (Widget* w, byte flags, EventHandler* ev)
{
	Widget* tmp;

	tmp = malloc (sizeof(Widget));
	tmp->flags = flags;
	tmp->on_event = ev;
	tmp->state = NULL;
	tmp->parent = w;
	tmp->next = NULL;
	tmp->child_list = NULL;
	tmp->on_event = ev;
	return tmp;
}

// Widgets are added as children of the widget on which they are placed. This makes it possible
// to iterate through the widgets recursively.
inline Widget*	CreateWidget (Widget* w, byte flags, EventHandler* ev)
{
	Widget* tmp;

// Create the new widget and fill it with basic settings.
// Put it in its place.
	if (w == NULL)
		w = BuildWidget (w,flags,ev);										// It's the topmost widget.
	else {
		if (w->child_list != NULL) {										// It's a new child.
			tmp = w->child_list;
			while (tmp->next != NULL)
				tmp = tmp->next;
			tmp->next = BuildWidget (w,flags,ev);
		} else
			w->child_list = BuildWidget (w,flags,ev);			// It's the first child.
	}
}

/////////////////////////////////////////////////////////////////////////////
// Widget helper functions.
//

// Puts a widget on top of the list.
inline void PutWidgetOnTop (Widget* w)
{
	assert (w->parent != NULL);

	Widget* tmp = w->parent->child_list;
// First connect the previous item to the next item.
	while (tmp->next != w)
		tmp = tmp->next;
	tmp->next = w->next;
// Set w as the first item.
	tmp = w->parent->child_list;
	w->next = tmp;
	w->parent->child_list = w;
}

static inline Widget* RecurseFindWidgetXY (const Widget* w, Point xy, Point realxy)
{
	Widget* result = NULL;
	Widget* tmp;
	Point		rxy;

	if (w->state & XY_ABSOLUTE != 0)
		rxy = w->xy
	else
		rxy = realxy + w->xy;

	if (InArea (rxy,xy) {
		tmp = w->child_list;
// Check if a child happens to be at xy.
		while ((tmp != NULL) && (result == NULL)) {
			result = RecurseFindWidgetXY(tmp,xy,rxy);
			tmp = tmp->next;
		}
		if (result == NULL)
			result = w;
	}
	return result;	
}

// Returns the topmost widget at xy.
Widget* FindWidgetXY (Widget* w, Point xy)
{
	return (RecurseFindWidgetXY (w,xy,{0,0,0,0});
}

inline void MarkForRedraw (const Widget* w)
{
	Widget* tmp = w;

	while (tmp != NULL) {
		tmp->flags |= REDRAW_MARK;
		tmp = tmp->parent;
	}
}

inline void HandleRedraw (const Widget* w)
{
	Widget*		tmp = w;
	Event			e;
	while (tmp != NULL) {
		if ((tmp->state & REDRAW_MARK) == true) && (tmp->child_list != NULL))
			HandleRedraw (tmp->child_list);
		tmp = tmp->next;
	}
	e = {E_REDRAW, 0, NULL};
	w->on_event(&e, w);
}
